# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTIBILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import bpy
from .op_cube import *
from .op_cylinder import *
from .op_sphere import *
from .op_pyramid import *

from .tools.turn_cylinder import *
from .qblock.qblock_prop import *


bl_info = {
    "name": "QBlocker",
    "author": "Balazs Szeleczki",
    "version": (0, 1, 4, 0),
    "description": "",
    "blender": (2, 80, 0),
    "location": "View3D > Add > QBlocker",
    "warning": "",
    "category": "Add Mesh",
    "wiki_url": "http://iiidea.cn/plug-ins/blender-plug"
}


class QBlockerAddonPreferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    mouse_enum : bpy.props.EnumProperty(
        name="select mouse",
        items=(
        ('LEFT', 'Left', ""),
        ('RIGHT', 'Right', "")
        ),
        default='LEFT'
    )

    axis_bool : bpy.props.BoolProperty(
        name="Show orientation axis",
        description="Show axis while moving mouse in view",
        default = True
    ) 

    object_ignorebehind : bpy.props.BoolProperty(
        name="Ignore objects behind grid",
        description="If the object behind the grid, the target will be calculated on the grid.",
        default = True
    ) 

    segs_int : bpy.props.IntProperty(
        name="Default segments",
        description="Default segments count for cylinder and sphere.",
        default = 16,
        min = 4,
        max = 128
    )

    text_size_int : bpy.props.IntProperty(
        name="Text size",
        description="Default segments count for cylinder and sphere.",
        default = 14,
        min = 4,
        max = 128
    )

    text_color : bpy.props.FloatVectorProperty(
        name="Text color",
        description="The closest snap point color.", 
        subtype='COLOR',
        default=[1.0, 1.0, 1.0, 1.0],
        size=4,
        min=0, max=1
    )

    hotkey_color : bpy.props.FloatVectorProperty(
        name="Hotkey color",
        description="The closest snap point color.", 
        subtype='COLOR',
        default=[1.0, 0.86, 0.0, 1.0],
        size=4,
        min=0, max=1
    )

    snap_dist : bpy.props.IntProperty(
        name="Snap Distance",
        description="Default distance between mouse and the vertices in.",
        default = 30,
        min = 5,
        max = 500
    )

    snap_dotsize : bpy.props.FloatProperty(
        name="Dot size factor",
        description="Default distance between mouse and the vertices in.",
        default = 1.0,
        min = 0.1,
        max = 10
    )

    snap_closestColor : bpy.props.FloatVectorProperty(
        name="Closest point color",
        description="The closest snap point color.", 
        subtype='COLOR',
        default=[1.0, 1.0, 1.0, 1.0],
        size=4,
        min=0, max=1
    )

    snap_pointsColor : bpy.props.FloatVectorProperty(
        name="Poly point color",
        description="The closest snap point color.", 
        subtype='COLOR',
        default=[1.0, 0.56, 0.0, 0.7],
        size=4,
        min=0, max=1
    )

    snap_gridpointsColor : bpy.props.FloatVectorProperty(
        name="Grid point color",
        description="The closest snap point color.", 
        subtype='COLOR',
        default=[0.12, 0.56, 1.0, 0.7],
        size=4,
        min=0, max=1
    )

    def draw(self, context):
        layout = self.layout
        general_box = layout.box()
        general_box.label(text="General settings")
        general_box.label(text="Active mouse button for the addon:")
        general_box.row().prop(self, "mouse_enum", expand=True)
        general_box.row().prop(self, "axis_bool")
        general_box.row().prop(self, "object_ignorebehind")
        
        general_box.row().prop(self, "text_size_int")
        general_box.row().prop(self, "text_color")
        general_box.row().prop(self, "hotkey_color")  


        snap_box = layout.box()
        snap_box.label(text="Snap settings")

        snap_box.row().prop(self, "snap_dist")
        snap_box.row().prop(self, "snap_dotsize")
        snap_box.row().prop(self, "snap_closestColor")
        snap_box.row().prop(self, "snap_pointsColor")
        snap_box.row().prop(self, "snap_gridpointsColor")

        object_box = layout.box()
        object_box.label(text="Object settings")
        object_box.row().prop(self, "segs_int", expand=True)


# right click menu button
class VIEW3D_MT_qbtools(bpy.types.Menu):
    bl_idname = "VIEW3D_MT_qbtools"
    bl_label = "QBlocker"

    def draw(self, context):
        layout = self.layout
        layout.operator("object.box_create", text="Plane / Cube", icon='MESH_CUBE')
        layout.operator("object.cylinder_create", text="Circle / Cylinder", icon='MESH_CYLINDER')
        layout.operator("object.sphere_create", text="Sphere", icon='MESH_UVSPHERE')
        #layout.operator("object.pyramid_create", text="Pyramid", icon='MESH_UVSPHERE')
        layout.separator()
        layout.operator("object.turn_to_cylinder", text="Turn into Cylinder", icon='MESH_UVSPHERE')

           
def qblocker_menu(self, context):
    self.layout.menu("VIEW3D_MT_qbtools")
    self.layout.separator()


classes = (
    BoxCreateOperator,
    CylinderCreateOperator,
    SphereCreateOperator,
    #PyramidCreateOperator,
    Turn_Sphere,
    VIEW3D_MT_qbtools,
    QBlockProperties,
    MESH_PT_QblockPanel
)


def register():
    for cls in classes:
        bpy.utils.register_class(cls)
    #bpy.types.VIEW3D_MT_object_context_menu.prepend(qblocker_menu)
    bpy.types.VIEW3D_MT_add.prepend(qblocker_menu)
    bpy.utils.register_class(QBlockerAddonPreferences)
    bpy.types.Mesh.qblock_props = bpy.props.PointerProperty(type=QBlockProperties)


def unregister():
    for cls in reversed(classes):
        bpy.utils.unregister_class(cls)
    #bpy.types.VIEW3D_MT_object_context_menu.remove(qblocker_menu)
    bpy.types.VIEW3D_MT_add.remove(qblocker_menu)
    bpy.utils.unregister_class(QBlockerAddonPreferences)

if __name__ == "__main__":
    register()
