import bpy
import mathutils
import bmesh


def SetSmoothFaces(bMesh):
    bMesh.use_auto_smooth = True
    for face in bMesh.polygons:
        face.use_smooth = True

def Create_ScaleMatrix(size):
    mat_scaleX = mathutils.Matrix.Scale(size[0], 4, (1.0, 0.0, 0.0))
    mat_scaleY = mathutils.Matrix.Scale(size[1], 4, (0.0, 1.0, 0.0))
    mat_scaleZ = mathutils.Matrix.Scale(size[2], 4, (0.0, 0.0, 1.0))
    return  (mat_scaleX @ mat_scaleY @ mat_scaleZ)

def update_QCube(self, context):
    bmeshNew = bmesh.new()
    mat_scale = Create_ScaleMatrix(self.size)
    if not self.is_centered:
        mat_scale = mathutils.Matrix.Translation( (0.0, 0.0, (self.size[2] / 2.0) )) @ mat_scale
    bmesh.ops.create_cube(bmeshNew,
                        size=1.0,
                        matrix=mat_scale,
                        calc_uvs=True)

    bmeshNew.to_mesh(self.id_data)
    bpy.context.view_layer.update()
    bmeshNew.free()

def update_QPlane(self, context):
    bmeshNew = bmesh.new()
    mat_scale = Create_ScaleMatrix(self.size)
    bmesh.ops.create_grid(bmeshNew,
                        x_segments=1,
                        y_segments=1,
                        size=0.5,
                        matrix=mat_scale,
                        calc_uvs=True)

    bmeshNew.to_mesh(self.id_data)
    bpy.context.view_layer.update()
    bmeshNew.free()

def update_QCylinder(self, context):
    bmeshNew = bmesh.new()
    mat_scale = Create_ScaleMatrix(self.aspect)
    if not self.is_centered:
        mat_scale = mathutils.Matrix.Translation( (0.0, 0.0, (self.depth / 2.0) )) @ mat_scale
    bmesh.ops.create_cone(bmeshNew,
                        segments=self.segments,
                        cap_ends=True,
                        cap_tris=False,
                        diameter1=self.radius,
                        diameter2=self.radius,
                        depth=self.depth,
                        matrix=mat_scale,
                        calc_uvs=True)

    bmeshNew.to_mesh(self.id_data)
    if self.is_smooth:
        self.id_data.auto_smooth_angle = 1.386
        SetSmoothFaces(self.id_data)
    bpy.context.view_layer.update()
    bmeshNew.free()

def update_QCircle(self, context):
    bmeshNew = bmesh.new()
    mat_scale = Create_ScaleMatrix(self.aspect)
    bmesh.ops.create_circle(bmeshNew,
                            segments=self.segments,
                            cap_ends=True,
                            cap_tris=False,
                            radius=self.radius,
                            matrix=mat_scale,
                            calc_uvs=True)
    bmeshNew.to_mesh(self.id_data)
    if self.is_smooth:
        self.id_data.auto_smooth_angle = 1.386
        SetSmoothFaces(self.id_data)
    bpy.context.view_layer.update()
    bmeshNew.free()

def update_QSphere(self, context):
    bmeshNew = bmesh.new()
    mat_scale = Create_ScaleMatrix(self.aspect)
    if not self.is_centered:
        mat_scale = mathutils.Matrix.Translation( (0.0, 0.0, (self.aspect[2]) * self.radius )) @ mat_scale
    bmesh.ops.create_uvsphere(bmeshNew,
                                u_segments=self.segments,
                                v_segments=int(self.segments / 2),
                                diameter=self.radius,
                                matrix=mat_scale,
                                calc_uvs=True)
    bmeshNew.to_mesh(self.id_data)  
    if self.is_smooth:
        self.id_data.auto_smooth_angle = 1.386
        SetSmoothFaces(self.id_data)
    bpy.context.view_layer.update()
    bmeshNew.free()

def draw_cube(self, context):
    layout = self.layout
    Qblock_Data = context.object.data.qblock_props
    col = layout.column(align=True)
    layout.prop(Qblock_Data, "is_centered", text = "Centered")
    col.prop(Qblock_Data, "size", text = "Scale")

def draw_plane(self, context):
    layout = self.layout
    Qblock_Data = context.object.data.qblock_props
    col = layout.column(align=True)
    col.prop(Qblock_Data, "size", text = "Scale")

def draw_cylinder(self, context):
    layout = self.layout
    Qblock_Data = context.object.data.qblock_props
    layout.prop(Qblock_Data, "is_centered", text = "Centered")
    layout.prop(Qblock_Data, "is_smooth", text = "Smooth")
    layout.prop(Qblock_Data, "radius", text = "Radius")
    layout.prop(Qblock_Data, "segments", text = "Segments")
    layout.prop(Qblock_Data, "depth", text = "Depth")
    col = layout.column(align=True)
    col.prop(Qblock_Data, "aspect", text = "Aspect")

def draw_circle(self, context):
    layout = self.layout
    Qblock_Data = context.object.data.qblock_props
    layout.prop(Qblock_Data, "radius", text = "Radius")
    layout.prop(Qblock_Data, "segments", text = "Segments")
    col = layout.column(align=True)
    col.prop(Qblock_Data, "aspect", text = "Aspect")

def draw_sphere(self, context):
    layout = self.layout
    Qblock_Data = context.object.data.qblock_props
    layout.prop(Qblock_Data, "is_centered", text = "Centered")
    layout.prop(Qblock_Data, "is_smooth", text = "Smooth")
    layout.prop(Qblock_Data, "radius", text = "Radius")
    layout.prop(Qblock_Data, "segments", text = "Segments")
    col = layout.column(align=True)
    col.prop(Qblock_Data, "aspect", text = "Aspect")


def update_QObject(self, context):
    Qblock_Data = context.object.data.qblock_props
    if Qblock_Data.qblockType == 'QCUBE':
        update_QCube(self, context)
    if Qblock_Data.qblockType == 'QPLANE':
        update_QPlane(self, context)
    if Qblock_Data.qblockType == 'QCYLINDER':
        update_QCylinder(self, context)
    if Qblock_Data.qblockType == 'QCIRCLE':
        update_QCircle(self, context)
    if Qblock_Data.qblockType == 'QSPHERE':
        update_QSphere(self, context)


class QBlockProperties(bpy.types.PropertyGroup):
    QBTypes = [
        ('NONE', "None", ""),
        ('QCUBE', "Cube", ""),
        ('QPLANE', "Plane", ""),
        ('QCYLINDER', "Cylinder", ""),
        ('QCIRCLE', "Circle", ""),
        ('QSPHERE', "Sphere", "")
    ]

    qblockType : bpy.props.EnumProperty(name="QBlock type", items=QBTypes, default='NONE', update = update_QObject)
    is_centered : bpy.props.BoolProperty(name="is_centered", update = update_QObject)
    is_smooth : bpy.props.BoolProperty(name="is_smooth", update = update_QObject)
    size : bpy.props.FloatVectorProperty(name="size", default=(1.0,1.0,1.0), precision=3, subtype='XYZ', unit='LENGTH', update = update_QObject)
    aspect : bpy.props.FloatVectorProperty(name="aspect", default=(1.0,1.0,1.0), precision=3, subtype='XYZ', update = update_QObject)
    radius: bpy.props.FloatProperty(name="radius", default=1.0, min=0, precision=3, unit='LENGTH', update = update_QObject)
    segments: bpy.props.IntProperty(name="segments", default=16, min=4, max=256,update = update_QObject)
    depth: bpy.props.FloatProperty(name="depth", default=1.0, precision=3, update = update_QObject)
    

class MESH_PT_QblockPanel(bpy.types.Panel):
    bl_idname = "DATA_PT_QblockPanel"
    bl_label = "QBlock Properties"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "data"

    @classmethod
    def poll(cls, context):
        return (context.object.type == 'MESH')
    
    def draw(self, context):
        layout = self.layout
        obj = context.object
        Qblock_Data = context.object.data.qblock_props
        layout.prop(Qblock_Data, "qblockType", text ="Type")
        if Qblock_Data.qblockType == 'QCUBE':
            draw_cube(self, context)
        if Qblock_Data.qblockType == 'QPLANE':
            draw_plane(self, context)
        if Qblock_Data.qblockType == 'QCYLINDER':
            draw_cylinder(self, context)
        if Qblock_Data.qblockType == 'QCIRCLE':
            draw_circle(self, context)
        if Qblock_Data.qblockType == 'QSPHERE':
            draw_sphere(self, context)



