import bpy
import gpu
import bgl
from gpu_extras.batch import batch_for_shader
import mathutils
import numpy
# import time
from .utilities.grid_utils import *
from .utilities.math_utils import *

redC = (0.984, 0.2, 0.318, 1.0)
greenC = (0.525, 0.824, 0.012, 1.0)
blueC = (0.157, 0.557, 0.988, 1.0)
vcolors = [greenC, greenC, redC, redC, blueC, blueC]

def QSpaceDrawHandler(qSpace, context):
    bgl.glLineWidth(2)
    pos = qSpace.wMatrix.translation
    mat = qSpace.wMatrix.to_3x3()
    vecX = pos + mat.col[0] 
    vecY = pos + mat.col[1] 
    vecZ = pos + mat.col[2] 

    coords = [pos, vecX, pos, vecY, pos, vecZ]
    shader = gpu.shader.from_builtin('3D_SMOOTH_COLOR')
    batch = batch_for_shader(shader, 'LINES', {"pos": coords, "color": vcolors})

    shader.bind()
    batch.draw(shader)


class CoordSysClass:
    operator = None
    addon_prefs = None
    context = None
    lastHitresult = (False, None, None, None, None, None)
    isGridhit = False
    isModifiedMesh = False
    mesh_data = None
    wMatrix = mathutils.Matrix()
    isPropAxis = True
    object_eval = None

    def __init__(self, _context, _op, _isAxis, _addon_prefs):
        self.addon_prefs = _addon_prefs
        self.operator = _op
        self.context = _context
        self.isPropAxis = _isAxis
        if _isAxis:
            args = (self, _context)
            self._handle = bpy.types.SpaceView3D.draw_handler_add(QSpaceDrawHandler, args, 'WINDOW', 'POST_VIEW')
        print("CoordSysClass created!")

    def CleanUp(self):
        if self.isPropAxis and self._handle:
            bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
        print("SnappingClass cleaned!")

    def ToggleAxis(self, _state):
        if self.isPropAxis:
            if _state:
                if not self._handle:
                    args = (self, self.context)
                    self._handle = bpy.types.SpaceView3D.draw_handler_add(QSpaceDrawHandler, args, 'WINDOW', 'POST_VIEW')
            else:
                if self._handle:
                    bpy.types.SpaceView3D.draw_handler_remove(self._handle, 'WINDOW')
                    self._handle = None


    def VecToMatrix(self, _vecZ, _pos):
        vecZ = _vecZ.normalized()
        if abs(vecZ[2]) == 1:
            vecY = vecZ.cross(mathutils.Vector((1.0, 0.0, 0.0)))
        else:
            vecY = vecZ.cross(mathutils.Vector((0.0, 0.0, -1.0)))
        vecY.normalize()
        vecX = vecY.cross(vecZ)
        vecX.normalize()

        matrix = mathutils.Matrix()
        matrix[0].xyz = vecX[0], vecY[0], vecZ[0]
        matrix[1].xyz = vecX[1], vecY[1], vecZ[1]
        matrix[2].xyz = vecX[2], vecY[2], vecZ[2]
        matrix.translation = _pos
        return matrix

    def Distance(self, pointA, pointB, _norm=numpy.linalg.norm):
        return _norm(pointA - pointB)

    # raycast into scene from mouse pos
    def HitScene(self, coord):
        scene = self.context.scene
        region = self.context.region
        rv3d = self.context.region_data
        view_vector = region_2d_to_vector_3d(region, rv3d, coord)
        ray_origin = region_2d_to_origin_3d(region, rv3d, coord)
        hitresult = scene.ray_cast(self.context.view_layer, ray_origin, view_vector)
        return hitresult

    def ResetResult(self):
        self.lastHitresult = (False, None, None, None, None, None)
        if self.isModifiedMesh:
            self.RemoveTempMesh()

    def RemoveTempMesh(self):
        self.object_eval.to_mesh_clear()
        self.mesh_data = None
        self.isModifiedMesh = False   

    def IsFrontOfGrid(self, coord, hitresult):
        if self.operator.object_ignorebehind:
            if hitresult[0]:
                scene = self.context.scene
                region = self.context.region
                rv3d = self.context.region_data
                ray_origin = region_2d_to_origin_3d(region, rv3d, coord)
                hitDist = self.Distance(hitresult[1], ray_origin)
                gridMat = self.GetActiveGridAlign(coord)
                gridHitDist = self.Distance(gridMat.to_translation(), ray_origin)
                if hitDist + 0.0001 < gridHitDist:
                    self.isGridhit = True
                    return True
        else:
             return True
        return False

    # main function
    def GetCoordSys(self, context, coord, isoriented):
        # time_start = time.time() # timetest
        if self.operator.toolstage != 0:
            self.operator.qObject.HideObject(True)
        hitresult = self.HitScene(coord)
        cSysMatrix = None
        # if object hit
        if hitresult[0] and hitresult[4].type == 'MESH' and self.IsFrontOfGrid(coord, hitresult):
            # if not the same object, then create evaluated mesh data
            if hitresult[4] != self.lastHitresult[4]: 
                # remove last mesh if that was modified 
                if self.isModifiedMesh:
                    self.RemoveTempMesh()  
                # check if hit modified object
                if len(hitresult[4].modifiers) != 0:
                    depsgraph = context.evaluated_depsgraph_get()
                    self.object_eval = hitresult[4].evaluated_get(depsgraph)
                    self.mesh_data = self.object_eval.to_mesh()
                    self.isModifiedMesh = True   
                else:
                    self.mesh_data = hitresult[4].data
            # Get matrix if oriented or axis aligned   
            if isoriented:
                cSysMatrix = self.GetOrientedAlign(hitresult)
            else:
                cSysMatrix = self.GetAxisAlign(hitresult)
            self.isGridhit = False
            self.lastHitresult = hitresult
        # if gridhit       
        else:
            if not self.isGridhit:
                self.ResetResult() 
                self.isGridhit = True
            cSysMatrix = self.GetActiveGridAlign(coord)
            self.lastHitresult = (False, None, None, None, None, None)

        if self.operator.toolstage != 0:
            self.operator.qObject.HideObject(False)

        self.wMatrix = cSysMatrix 
        return cSysMatrix

    # Get space align (wqrking plane or world grid)
    def GetActiveGridAlign(self, coord):
        cSysMatrix = None
        if self.operator.isWorkingPlane:
            cSysMatrix = self.GetWPlaneAlign(coord, self.operator.workingplane.matrix)
        else:
            cSysMatrix = self.GetGridAlign(coord)
        return cSysMatrix

    # GRID ALIGNED
    def GetGridAlign(self, coord):
        # get view ray
        scene = self.context.scene
        region = self.context.region
        rv3d = self.context.region_data
        view_vector = region_2d_to_vector_3d(region, rv3d, coord)
        ray_origin = region_2d_to_origin_3d(region, rv3d, coord)

        # check grid and collide with view ray
        grid_vector = GetGridVector(self.context)
        hitpoint = LinePlaneCollision(view_vector, ray_origin, (0.0, 0.0, 0.0), grid_vector)

        # create matrix
        grid_vector = mathutils.Vector(grid_vector)
        ret_matrix = self.VecToMatrix(grid_vector, hitpoint)
        return ret_matrix  

    # WPLANE ALIGNED
    def GetWPlaneAlign(self, coord, _matrix):
        # get view ray
        scene = self.context.scene
        region = self.context.region
        rv3d = self.context.region_data
        view_vector = region_2d_to_vector_3d(region, rv3d, coord)
        ray_origin = region_2d_to_origin_3d(region, rv3d, coord)

        grid_vector = _matrix.col[2].xyz
        hitpoint = LinePlaneCollision(view_vector, ray_origin, _matrix.translation, grid_vector)

        # create matrix
        ret_matrix = _matrix.copy()
        ret_matrix.translation = hitpoint
        return ret_matrix  

    # AXIS ALIGNED
    def GetAxisAlign(self, _hitresult):
        return self.VecToMatrix(_hitresult[2], _hitresult[1])

    # ORIENTED ALIGN
    def GetOrientedAlign(self, _hitresult):
        verts = self.GetTargetPolyVerts(_hitresult)

        if _hitresult[4] == self.lastHitresult[4] and _hitresult[2] == self.lastHitresult[2]:
                self.wMatrix.translation = _hitresult[1]
                return self.wMatrix
        else:
            # create matrix from face normal
            matrix = self.VecToMatrix(_hitresult[2], _hitresult[1])
            mat_inv = matrix.inverted()

            # transform verts to poly space
            vertsL = []
            for v in verts:
                v2 = mat_inv @ v
                v2[2] = 0.0
                vertsL.append(v2)

            # calc best rotation
            verts2DL = [(p[0], p[1]) for p in vertsL]
            bboxangle = mathutils.geometry.box_fit_2d(verts2DL)
            mat_rot = mathutils.Matrix.Rotation(-bboxangle, 4, 'Z')
            ret_matrix = matrix @ mat_rot   
            return ret_matrix

    # get vertices from polygon in world space
    def GetTargetPolyVerts(self, _hitresult):
        verts = []
        meshFace = self.mesh_data.polygons[_hitresult[3]]
        matrix = _hitresult[4].matrix_world.copy()
        for v in meshFace.vertices:
            pos = _hitresult[4].matrix_world @ (self.mesh_data.vertices[v].co)
            verts.append(pos)
        return verts